"""Velocity distribution function to be used by the integrand class `a_vdf`.

One of the integrands available for use in the Gordeyev integral.
"""

from abc import ABC, abstractmethod

import numpy as np
import scipy.constants as const
import scipy.special as sps

from inscar import config


class Vdf(ABC):
    """Base class for a VDF object.

    Parameters
    ----------
    ABC:
        Abstract base class that all Vdf objects inherit from
    """

    @abstractmethod
    def normalize(self) -> None:
        """Calculate the normalization for the VDF."""

    @abstractmethod
    def f_0(self) -> np.ndarray:
        """Return the values along the velocity axis of a VDF.

        Returns
        -------
        np.ndarray
            1D array with the VDF values at the sampled points
        """


class VdfMaxwell(Vdf):
    """Create an object that make Maxwellian distribution functions.

    Parameters
    ----------
    VDF: ABC
        Abstract base class to make VDF objects
    """

    def __init__(self, params: config.Parameters, particle: config.Particle):
        """Initialize VDF parameters.

        Parameters
        ----------
        params : Parameters
            Parameters object with the parameters of the simulation.
        particle : Particle
            Particle object with the parameters of the particle.
        """
        self.params = params
        self.particle = particle
        self.normalize()

    def normalize(self) -> None:
        self.A = (
            2 * np.pi * self.particle.temperature * const.k / self.particle.mass
        ) ** (-3 / 2)

    def f_0(self) -> np.ndarray:
        return self.A * np.exp(
            -self.particle.velocity_axis**2
            / (2 * self.particle.temperature * const.k / self.particle.mass)
        )


class VdfKappa(Vdf):
    """Create an object that make kappa distribution functions.

    Kappa VDF used in Gordeyev paper by Mace (2003).

    """

    def __init__(self, params: config.Parameters, particle: config.Particle):
        """Initialize VDF parameters.

        Parameters
        ----------
        params : Parameters
            Parameters object with the parameters of the simulation.
        particle : Particle
            Particle object with the parameters of the particle.
        """
        self.params = params
        self.particle = particle
        self.normalize()

    def normalize(self) -> None:
        self.theta_2 = (
            2
            * ((self.particle.kappa - 3 / 2) / self.particle.kappa)
            * self.particle.temperature
            * const.k
            / self.particle.mass
        )
        self.A = (
            (np.pi * self.particle.kappa * self.theta_2) ** (-3 / 2)
            * sps.gamma(self.particle.kappa + 1)
            / sps.gamma(self.particle.kappa - 1 / 2)
        )

    def f_0(self) -> np.ndarray:
        return self.A * (
            1 + self.particle.velocity_axis**2 / (self.particle.kappa * self.theta_2)
        ) ** (-self.particle.kappa - 1)



"""ISR Spectrum."""

from importlib_metadata import version

from inscar.config import *  # noqa: F401, F403
from inscar.integrand_functions import *  # noqa: F401, F403
from inscar.spectrum_calculation import SpectrumCalculation  # noqa: F401
from inscar.vdfs import *  # noqa: F401, F403

__version__ = version(__package__)



"""Objects used to configure the physical parameters of the system."""
from typing import Tuple, Union

import attr
import numpy as np
import scipy.constants as const


def is_odd(_, attribute, value):
    """Verify that a value is odd."""
    if value % 2 == 0:
        raise ValueError(f"{attribute} must be odd")


def is_positive(_, attribute, value):
    """Verify that a value is positive."""
    if not isinstance(value, (int, float)):
        raise ValueError(f"{attribute} must be a positive number")
    if value <= 0:
        raise ValueError(f"{attribute} must be positive")


def is_nonnegative(_, attribute, value):
    """Verify that a value is non-negative."""
    if not isinstance(value, (int, float)):
        raise ValueError(f"{attribute} must be a non-negative number")
    if value < 0:
        raise ValueError(f"{attribute} must be non-negative")


def is_range_tuple(_, attribute, value):
    """Verify that a value is a tuple of two numbers."""
    if not isinstance(value, tuple):
        raise ValueError(f"{attribute} must be a tuple")
    if len(value) != 2:
        raise ValueError(f"{attribute} must be a tuple of length 2")
    if isinstance(value[0], (int, float)) and isinstance(value[1], (int, float)):
        if value[0] >= value[1]:
            raise ValueError(f"{attribute} must be a tuple of increasing values")
    else:
        raise ValueError(f"{attribute} must be a tuple of int/floats")


def to_radians(value) -> float:
    """Convert degrees to radians."""
    return value * np.pi / 180


@attr.s(auto_attribs=True)
class Particle:
    """Object used to configure the physical parameters of a particle."""

    gordeyev_upper_lim: Union[float, int] = attr.ib(
        default=1.5e-4,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    gordeyev_size: int = attr.ib(
        default=int(8e4 + 1),
        validator=is_odd,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=int,
    )
    gordeyev_exp: int = attr.ib(
        default=3,
        validator=is_odd,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=int,
    )
    velocity_upper_lim: Union[float, int] = attr.ib(
        default=6e6,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    velocity_size: int = attr.ib(
        default=int(4e4 + 1),
        validator=is_odd,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=int,
    )
    velocity_exp: int = attr.ib(
        default=3,
        validator=is_odd,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=int,
    )
    temperature: float = attr.ib(
        default=5000,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    collision_frequency: float = attr.ib(
        default=0,
        validator=is_nonnegative,
        on_setattr=attr.setters.validate,
    )
    mass: float = attr.ib(
        default=const.m_e,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    number_density: float = attr.ib(
        default=2e11,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    kappa: float = attr.ib(
        default=20,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )

    @property
    def velocity_axis(self) -> np.ndarray:
        """Return the axis for the velocity integral."""
        return (
            np.linspace(
                0,
                self.velocity_upper_lim ** (1 / self.velocity_exp),
                self.velocity_size,
            )
        ) ** self.velocity_exp

    @property
    def gordeyev_axis(self) -> np.ndarray:
        """Return the axis for the Gordeyev integral."""
        return (
            np.linspace(
                0,
                self.gordeyev_upper_lim ** (1 / self.gordeyev_exp),
                self.gordeyev_size,
                dtype=np.double,
            )
        ) ** self.gordeyev_exp


@attr.s(auto_attribs=True)
class Parameters:
    """Object used to configure the physical parameters of the system."""

    radar_frequency: float = attr.ib(
        default=430e6,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    frequency_range: Tuple = attr.ib(
        default=(-2e6, 2e6),
        validator=is_range_tuple,
        on_setattr=attr.setters.validate,
    )
    frequency_size: int = attr.ib(
        default=int(1e4 + 1),
        validator=is_odd,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=int,
    )
    frequency_exp: int = attr.ib(
        default=1,
        validator=is_odd,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=int,
    )
    magnetic_field_strength: float = attr.ib(
        default=35000e-9,
        validator=is_positive,
        on_setattr=attr.setters.validate,
    )
    aspect_angle: Union[int, float] = attr.ib(
        default=45,
        validator=attr.validators.instance_of((int, float)),
        on_setattr=[attr.setters.convert, attr.setters.validate],
        converter=to_radians,
    )

    @property
    def linear_frequency(self) -> np.ndarray:
        """Return the linear frequency used for the power spectrum."""
        return (
            np.linspace(
                self.frequency_range[0], self.frequency_range[1], self.frequency_size
            )
            / self.frequency_range[1]
        ) ** self.frequency_exp * self.frequency_range[1]

    @property
    def angular_frequency(self) -> np.ndarray:
        """Return the angular frequency used for the power spectrum."""
        return self.linear_frequency * 2 * np.pi

    @property
    def radar_wavenumber(self) -> float:
        """Return the radar wave number."""
        return -2 * self.radar_frequency * 2 * np.pi / const.c



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['inscar', 'inscar.experimental']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.4.0,<22.0.0',
 'importlib-metadata>=4.11.4,<5.0.0',
 'matplotlib>=3.4.1,<4.0.0',
 'numba>=0.55.1,<0.56.0',
 'numpy>=1.20.2,<2.0.0',
 'scipy>=1.6.2,<2.0.0']

setup_kwargs = {
    'name': 'inscar',
    'version': '3.0.2',
    'description': 'Calculate an incoherent scatter spectrum with arbitrary isotropic electron velocity distributions and radar pointing at oblique angles to the magnetic field',
    'long_description': 'inscar\n======\n\n    INcoherent SCAtter Radar spectrum\n\n|PyPI| |Status| |Python Version| |License| |Read the Docs| |Tests| |Codecov| |DOI|\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/inscar.svg\n   :target: https://pypi.org/project/inscar/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/inscar.svg\n   :target: https://pypi.org/project/inscar/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/inscar\n   :target: https://pypi.org/project/inscar\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/badge/License-MIT-yellow.svg\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/inscar/latest.svg?label=Read%20the%20Docs\n   :target: https://inscar.readthedocs.io/\n   :alt: Read the documentation at https://ncdump-rich.readthedocs.io/\n.. |Tests| image:: https://github.com/engeir/inscar/workflows/Tests/badge.svg\n   :target: https://github.com/engeir/inscar/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/engeir/inscar/branch/master/graph/badge.svg?token=P8S18UILSB\n   :target: https://codecov.io/gh/engeir/inscar\n   :alt: Codecov\n.. |DOI| image:: https://zenodo.org/badge/233043566.svg\n   :target: https://zenodo.org/badge/latestdoi/233043566\n   :alt: pre-commit\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n.. [![release](https://img.shields.io/github/release/engeir/inscar.svg)](https://github.com/engeir/inscar/releases/latest)\n.. ![CodeQL](https://github.com/engeir/inscar/workflows/CodeQL/badge.svg)\n\n.. image:: ./img/normal_is_spectra.png\n\nInfo\n----\n\nCalculates an incoherent scatter radar spectrum based on the theory presented in\n`Hagfors (1961)`_ and `Mace (2003)`_.\n\nInstalling\n----------\n\nYou can install *inscar* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install inscar\n\nUsage\n-----\n\nPlease see the `Modules Reference <Modules_>`_ for details.\n\nNumba\n^^^^^\n\nFaster integration is accomplished by computing in parallel. This is\naccomplished using `numba`.\n\nPhysical environment\n^^^^^^^^^^^^^^^^^^^^\n\nThe available plasma parameters that can be changed within the program are\n\n.. code:: text\n\n    === Input parameters ===\n    K_RADAR -- Radar wavenumber (= -4pi(radar frequency)/(speed of light)) [m^(-1)]\n    B -- Magnetic field strength [T]\n    MI -- Ion mass in atomic mass units [u]\n    NE -- Electron number density [m^(-3)]\n    NU_E -- Electron collision frequency [Hz]\n    NU_I -- Ion collision frequency [Hz]\n    T_E -- Electron temperature [K]\n    T_I -- Ion temperature [K]\n    T_ES -- Temperature of suprathermal electrons in the gauss_shell VDF [K] (no longer in use)\n    THETA -- Aspect angle [1]\n    Z -- Height used for calculated distribution [100, 599] [km]\n    mat_file -- Time of day for calculated distribution\n    pitch_angle -- Pitch angle for calculated distribution\n\nCalculation method\n^^^^^^^^^^^^^^^^^^\n\nThe program support different methods of calculating the spectrum, based on how you\nassume the particles to be distributed. This includes a Maxwellian distribution and a\nkappa distribution, in addition to any arbitrary isotropic distribution.\n\n.. _Hagfors (1961): https://agupubs.onlinelibrary.wiley.com/doi/epdf/10.1029/JZ066i006p01699\n.. _Mace (2003): https://aip.scitation.org/doi/pdf/10.1063/1.1570828\n.. _PyPI: https://pypi.org/\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Modules: https://inscar.readthedocs.io/en/latest/modules.html\n',
    'author': 'engeir',
    'author_email': 'eirroleng@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
